import gpmp.num as gnp
import gpmp as gp
import scipy.stats as stats

def predict_with_bayesian_confidence_intervals(model, xi, zi, xt, alpha=0.05):
    """
    Compute Bayesian confidence intervals from posterior mean and variance.

    Parameters
    ----------
    zpm : array_like
        Posterior mean, 1D array of shape (n,).
    zpv : array_like
        Posterior variance, 1D array of shape (n,).
    alpha : float, optional
        Significance level for the confidence interval (default is 0.05).

    Returns
    -------
    lower_bounds : array-like
        Lower bounds of the confidence intervals.
    upper_bounds : array_like
        Upper bounds of the confidence intervals.
    """

    zpm, zpv = model.predict(xi, zi, xt)
        
    z_score = stats.norm.ppf(1 - alpha / 2)

    std_dev = gnp.sqrt(zpv)

    lower_bounds = zpm - z_score * std_dev
    upper_bounds = zpm + z_score * std_dev

    return zpm, zpv, lower_bounds, upper_bounds

def predict_with_jacknife_confidence_intervals(model, xi, zi, xt, alpha=0.05):
    



"""GP Conditional Sample Paths

This script constructs a GP model using a Matérn kernel and a constant
mean function. The script generates sample paths from the GP prior,
and then generates conditional sample paths given the observed data.

Copyright (c) 2022-2023, CentraleSupelec
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
License: GPLv3 (see LICENSE)

"""

import math
import numpy as np
import gpmp.num as gnp
import gpmp as gp
import gpmp.misc.plotutils as plotutils


def generate_data():
    nt = 200
    xt = np.linspace(-1, 1, nt).reshape(-1, 1)
    zt = gp.misc.testfunctions.twobumps(xt)

    ind = [10, 45, 100, 130, 155]
    xi = xt[ind]
    zi = zt[ind]

    return xt, zt, xi, zi, ind


def kernel(x, y, covparam, pairwise=False):
    p = 2
    return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def visualization(xt, zt, zpsim, xi, zi, zpm, zpv):
    fig = plotutils.Figure(isinteractive=True)
    fig.plot(xt, zt, 'C2', linewidth=1, label='truth')
    fig.plot(xt, zpsim[:, 1], 'C0', linewidth=1, label='posterior sample paths')
    fig.plot(xt, zpsim[:, 1:], 'C0', linewidth=1)
    fig.plot(xi, zi, 'rs')
    fig.plotgp(xt, zpm, zpv)
    fig.title('Conditional sample paths')
    fig.legend()
    fig.show()


def main():
    xt, zt, xi, zi, xi_ind = generate_data()

    mean = constant_mean
    meanparam = None
    covparam = gnp.array([math.log(0.5 ** 2), math.log(1 / .7)])
    model = gp.core.Model(mean, kernel, meanparam, covparam)

    n_samplepaths = 6
    zsim = model.sample_paths(xt, n_samplepaths, method='chol')

    zpm, zpv, lambda_t = model.predict(xi, zi, xt, return_lambdas=True)
    zpsim = model.conditional_sample_paths(zsim, xi_ind, zi, gnp.arange(xt.shape[0]), lambda_t)

    visualization(xt, zt, zpsim, xi, zi, zpm, zpv)


if __name__ == "__main__":
    main()



"""
Plot and optimize the restricted negative log-likelihood

Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2024, CentraleSupelec
License: GPLv3 (see LICENSE)
"""

import gpmp.num as gnp
import gpmp as gp
import matplotlib.pyplot as plt


def generate_data():
    """
    Data generation.

    Returns
    -------
    tuple
        (xt, zt): target data
        (xi, zi): input dataset
    """
    dim = 1
    nt = 200
    box = [[-1], [1]]
    xt = gp.misc.designs.regulargrid(dim, nt, box)
    zt = gp.misc.testfunctions.twobumps(xt)

    ni = 6
    xi = gp.misc.designs.ldrandunif(dim, ni, box)
    zi = gp.misc.testfunctions.twobumps(xi)

    return xt, zt, xi, zi


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def kernel(x, y, covparam, pairwise=False):
    p = 3
    return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)


def visualize_results(xt, zt, xi, zi, zpm, zpv):
    """
    Visualize the results using gp.misc.plotutils (a matplotlib wrapper).

    Parameters
    ----------
    xt : numpy.ndarray
        Target x values
    zt : numpy.ndarray
        Target z values
    xi : numpy.ndarray
        Input x values
    zi : numpy.ndarray
        Input z values
    zpm : numpy.ndarray
        Posterior mean
    zpv : numpy.ndarray
        Posterior variance
    """
    fig = gp.misc.plotutils.Figure(isinteractive=True)
    fig.plot(xt, zt, "k", linewidth=1, linestyle=(0, (5, 5)))
    fig.plotdata(xi, zi)
    fig.plotgp(xt, zpm, zpv, colorscheme="simple")
    fig.xylabels("$x$", "$z$")
    fig.title("Posterior GP with parameters selected by ReML")
    fig.show(xlim=[-1.0, 1.0])


def main():
    xt, zt, xi, zi = generate_data()

    meanparam0 = None
    covparam0 = None
    model = gp.core.Model(constant_mean, kernel, meanparam0, covparam0)

    # Parameter initial guess
    covparam0 = gp.kernel.anisotropic_parameters_initial_guess(model, xi, zi)

    # selection criterion
    nlrl, dnlrl = gp.kernel.make_selection_criterion_with_gradient(
        model, gp.kernel.negative_log_restricted_likelihood, xi, zi
    )

    covparam_reml, info = gp.kernel.autoselect_parameters(
        covparam0, nlrl, dnlrl, silent=False, info=True
    )

    model.covparam = gnp.asarray(covparam_reml)
    info["covparam0"] = covparam0
    info["covparam"] = covparam_reml
    info["selection_criterion"] = nlrl

    gp.misc.modeldiagnosis.diag(model, info, xi, zi)

    # Prediction
    zpm, zpv = model.predict(xi, zi, xt)

    # Visualization
    print("\nVisualization")
    print("-------------")
    visualize_results(xt, zt, xi, zi, zpm, zpv)

    zloom, zloov, eloo = model.loo(xi, zi)
    gp.misc.plotutils.plot_loo(zi, zloom, zloov)


if __name__ == "__main__":
    main()



## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2023-2024, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------
import gpmp.num as gnp




def crps_gaussian(mu, sigma, z):
    """
    Compute the CRPS for the Gaussian case.

    Parameters
    ----------
    mu : ndarray, shape (n,)
        Mean values of the Gaussian distribution.
    sigma : ndarray, shape (n,)
        Standard deviations of the Gaussian distribution.
    z : ndarray, shape (n,)
        Observations.

    Returns
    -------
    crps : ndarray, shape (n,)
        The computed CRPS values for each element in mu, sigma, and z.
    """
    mu = gnp.asarray(mu)
    sigma = gnp.asarray(sigma)
    z = gnp.asarray(z)
    
    t = (z - mu) / sigma
    term1 = t * (2 * gnp.normal.cdf(t) - 1)
    term2 = 2 * gnp.normal.pdf(t)
    term3 = 1 / gnp.sqrt(gnp.pi)
    crps = sigma * (term1 + term2 - term3)
    return crps


def h1(t):
    return t * gnp.normal.cdf(t) + gnp.normal.pdf(t)


def ei1_up(mu, sigma, z):
    """
    Compute the expected improvement EI1_up(P, z) for a Gaussian distribution.

    Parameters:
    -----------
    mu: float or np.ndarray
        The mean(s) of the Gaussian distribution(s).
    sigma: float or np.ndarray
        The standard deviation(s) of the Gaussian distribution(s).
    z: float or np.ndarray
        The input value(s).

    Returns:
    --------
    result: float or np.ndarray
        The EI1_up value(s) for the given Gaussian distribution(s).
    """
    t = (mu - z) / sigma
    return sigma * h1(t)


def ei2_up(mu, sigma, z):
    """
    Compute the expected improvement EI2_up(P, z) for a Gaussian distribution.

    Parameters:
    -----------
    mu: float or np.ndarray
        The mean(s) of the Gaussian distribution(s).
    sigma: float or np.ndarray
        The standard deviation(s) of the Gaussian distribution(s).
    z: float or np.ndarray
        The input value(s).

    Returns:
    --------
    result: float or np.ndarray
        The EI2_up(P, z) value(s) for the given Gaussian distribution(s).
    """
    t = (mu - z) / sigma
    if gnp.isscalar(t):
        t = gnp.array([t])
    delta_2_t = gnp.hstack((t.reshape(-1, 1), gnp.zeros((t.shape[0], 1))))

    D_2 = gnp.array([[-1, 0], [-1, 1]])

    term1 = 2.0 * t * gnp.multivariate_normal.cdf(
        delta_2_t, mean=gnp.zeros(2), cov=gnp.matmul(D_2, D_2.T)
    )
    term2 = 2.0 * gnp.normal.pdf(t) * gnp.normal.cdf(-t)
    term3 = 1.0 / gnp.sqrt(gnp.pi) * gnp.normal.cdf(t, loc=0.0, scale=gnp.sqrt(0.5))

    return sigma * (term1 + term2 + term3)


def h1(t):
    return t * gnp.normal.cdf(t) + gnp.normal.pdf(t)


def tcrps_gaussian(mu, sigma, z, a=-gnp.inf, b=gnp.inf):
    """
    Compute S_{a, b}^\tCRPS (P, z) for Gaussian distributions.

    Parameters:
    -----------
    mu: float or np.ndarray
        The mean(s) of the Gaussian distribution(s).
    sigma: float or np.ndarray
        The standard deviation(s) of the Gaussian distribution(s).
    z: float or np.ndarray
        The observed values.
    a: float or np.ndarray, optional
        The lower bound of the interval, default = -inf.
    b: float or np.ndarray, optional
        The upper bound of the interval, default = +inf.

    Returns:
    --------
    result: float or np.ndarray
        The S_{a, b}^\tCRPS (P, z) value(s) for the given Gaussian distribution(s).
    """
    mu = gnp.asarray(mu)
    sigma = gnp.asarray(sigma)
    z = gnp.asarray(z)
    a = gnp.asarray(a)
    b = gnp.asarray(b)
    if gnp.isfinite(a) and gnp.isfinite(b):
        term1 = gnp.maximum(gnp.minimum(b, z) - a, 0.0)
        term2 = ei2_up(mu, sigma, b) - ei2_up(mu, sigma, a)
        term3 = -2 * gnp.where(
            z <= b, ei1_up(mu, sigma, b) - ei1_up(mu, sigma, gnp.maximum(a, z)), 0
        )
        return term1 + term2 + term3
    elif ~gnp.isfinite(a) and gnp.isfinite(b):
        term1 = gnp.minimum(b, z)
        term2 = ei2_up(mu, sigma, b) - (mu + sigma / gnp.sqrt(gnp.pi))
        term3 = -2 * gnp.where(z <= b, ei1_up(mu, sigma, b) - ei1_up(mu, sigma, z), 0)
        return term1 + term2 + term3
    elif gnp.isfinite(a) and ~gnp.isfinite(b):
        return tcrps_gaussian(-mu, sigma, -z, a=-gnp.inf, b=-a)
    else:
        return crps_gaussian(mu, sigma, z)



import unittest
import gpmp as gp
import gpmp.num as gnp    

class TestScoringRules(unittest.TestCase):
    def test_tcrps_01(self):
        mu = 4.0
        sigma = gnp.sqrt(2)
        a = -gnp.inf
        b = 3.0
        z = 3.5
        x = gnp.to_scalar(gp.misc.scoringrules.tcrps_gaussian(mu, sigma, z, a, b))
        self.assertAlmostEqual(x, 0.02661950759116)

    def test_tcrps_02(self):
        mu = 1.86
        sigma = gnp.sqrt(0.8)
        a = 1.5
        b = 2.3
        z = 1.64
        x = gnp.to_scalar(gp.misc.scoringrules.tcrps_gaussian(mu, sigma, z, a, b))
        self.assertAlmostEqual(x, 0.159017709237)
        
    def test_tcrps_03(self):
        mu = -1.7
        sigma = gnp.sqrt(0.1)
        a = -1.0
        b = gnp.inf
        z = -1.5
        x = gnp.to_scalar(gp.misc.scoringrules.tcrps_gaussian(mu, sigma, z, a, b))
        self.assertAlmostEqual(x, 1.0475838916335078e-05)

    def test_tcrps_04(self):
        mu = 0.0 
        sigma = 1.0
        a = -1.0
        b = 1.0
        z = [1.64, 0.0, 0.5, -1.5]
        x = gp.misc.scoringrules.tcrps_gaussian(mu, sigma, z, a, b)
        x_expected = gnp.array([0.5879712039755658, 0.21922482360305862, 0.3169333776028054, 0.5879712039755658])
        self.assertAlmostEqual(gnp.to_scalar(gnp.norm(x - x_expected)), 0.)
    
        
if __name__ == "__main__":
    unittest.main()



# coding: utf-8
## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------
import math
import numpy as np


def twobumps(x):
    """
    Computes the response Z of the TwoBumps function at X.

    The TwoBumps function is defined as:

       TwoBumps(x) = - (0.7x + sin(5x + 1) + 0.1 sin(10x))

    Parameters
    ----------
    x : numpy.ndarray
        Input array of shape (n,)

    Returns
    -------
    numpy.ndarray
        Output array of shape (n,)
    """
    z = -(0.7 * x + (np.sin(5 * x + 1)) + 0.1 * (np.sin(10 * x)))
    return z.reshape([-1])


def wave(x):
    """Computes the Wave function.
    
    The Wave function is a popular test function used in optimization problems. It has multiple local minima and a single global minimum. It is defined as follows:
    
    .. math::
        f(x) = e^{1.8(x_1 + x_2)} + 3x_1 + 6x_2^2 + 3\sin(4\pi x_1)
        
    where x is a 2-dimensional numpy array in the range [-1, 1] x [-1, 1].
    
    Parameters
    ----------
    x : numpy.ndarray
        A 2-dimensional numpy array containing the input values of the function. The shape of the array should be (n, 2), where n is the number of input points.
        
    Returns
    -------
    numpy.ndarray
        A 1-dimensional numpy array containing the output values of the function. The shape of the array is (n,).
    """    
    z = (
        np.exp(1.8 * (x[:, 0] + x[:, 1]))
        + 3 * x[:, 1]
        + 6 * x[:, 1] ** 2
        + 3 * np.sin(4 * np.pi * x[:, 0])
    )

    return z


def braninhoo(x):
    """
    The Branin-Hoo function is a classical test function for global optimization algorithms, which belongs to the well-known Dixon-Szego test set. It is usually minimized over [-5; 10] x [0; 15].

    Parameters
    ----------
    x : numpy.array
        2D array of shape (n, 2) where each row represents a point in the 2D space to evaluate the function.

    Returns
    -------
    numpy.array
        A 1D array of shape (n,) containing the Branin-Hoo function values evaluated at the input points.

    Notes
    -----
    .. [1] Branin, F. H. and Hoo, S. K. (1972), A Method for Finding Multiple
        Extrema of a Function of n Variables, in Numerical methods of
        Nonlinear Optimization (F. A. Lootsma, editor, Academic Press,
        London), 231-237.

    .. [2] Dixon L.C.W., Szego G.P., Towards Global Optimization 2, North-
        Holland, Amsterdam, The Netherlands (1978)

    .. [3] Surjanovic, S. and Bingham D. (2013), Branin Function,
        https://www.sfu.ca/~ssurjano/Code/braninm.html
    """
    a = 5.1 / (4 * math.pi**2)
    b = 5 / math.pi
    c = 10 * (1 - 1 / (8 * math.pi))

    z = (x[:, 1] - a * x[:, 0] ** 2 + b * x[:, 0] - 6) ** 2 + c * np.cos(x[:, 0]) + 10

    return z


def hartmann4(x):
    """Hartmann 4-dimensional function [1, 2]

    The 4-dimensional Hartmann function is a multimodal function defined
    on the unit hypercube (i.e., xi in (0, 1), for all i = 1, ..., 4). It is
    commonly used as a test problem in global optimization.

    Parameters
    ----------
    x : numpy.ndarray
        An array of shape (n_samples, 4) containing the input points.

    Returns
    -------
    numpy.ndarray
        An array of shape (n_samples,) containing the function values at
        the input points.

    Notes
    -----
    The Hartmann 4-dimensional function has 6 local minima and one
    global minimum. The global minimum is located at x* = [0.20169,
    0.15001, 0.47687, 0.27533] and has a function value of f(x*) = -3.86278.
    The function is generally considered to be difficult to optimize.

    .. [1] Dixon, L. C. W., & Szego, G. P. (1978). The global optimization
        problem: an introduction. Towards global optimization, 2, 1-15.

    .. [2] Picheny, V., Wagner, T., & Ginsbourger, D. (2012). A benchmark
        of kriging-based infill criteria for noisy optimization.
        Based on https://www.sfu.ca/\~ssurjano/hart6.html

    Authors: Sonja Surjanovic and Derek Bingham, Simon Fraser University

    Original copyright notice: Copyright 2013. Derek Bingham, Simon Fraser University.
    """
    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    A = np.array(
        [
            [10, 3, 17, 3.5, 1.7, 8],
            [0.05, 10, 17, 0.1, 8, 14],
            [3, 3.5, 1.7, 10, 17, 8],
            [17, 8, 0.05, 10, 0.1, 14],
        ]
    )
    P = 1e-4 * np.array(
        [
            [1312, 1696, 5569, 124, 8283, 5886],
            [2329, 4135, 8307, 3736, 1004, 9991],
            [2348, 1451, 3522, 2883, 3047, 6650],
            [4047, 8828, 8732, 5743, 1091, 381],
        ]
    )

    outer = 0
    for ii in range(4):
        inner = 0
        for jj in range(4):
            xj = x[:, jj]
            Aij = A[ii, jj]
            Pij = P[ii, jj]
            inner = inner + Aij * (xj - Pij) ** 2
        new = alpha[ii] * np.exp(-inner)
        outer = outer + new

    z = (1.1 - outer) / 0.839

    return z


def hartmann6(x):
    """Hartmann 6-dimensional function [1]

    The 6-dimensional Hartmann function has 6 local minima and a
    a global minimum f(x*) = -3.32237.

    .. math::

        f(x) = - \\sum_{i=1}^4 \\alpha_i \\exp \\bigl(-\\sum_{j=1}^6 A_{ij}(x_j - P_{ij})^2 \\bigr) 

    where :math:`x_i \in (0, 1)` for all :math:`i = 1, \ldots, 6`.

    Parameters
    ----------
    x : numpy.ndarray
        An array of shape (n, 6) containing the input values.

    Returns
    -------
    numpy.ndarray
        An array of shape (n,) containing the function values.

    Notes
    -----
    
    .. [1] Dixon, L. C. W., & Szego, G. P. (1978). The global
        optimization problem: an introduction. Towards global
        optimization, 2, 1-15.

    .. [2] Picheny, V., Wagner, T., & Ginsbourger, D. (2012). A benchmark
        of kriging-based infill criteria for noisy optimization.
        Based on https://www.sfu.ca/~ssurjano/hart6.html

    Authors: Sonja Surjanovic and Derek Bingham, Simon Fraser University

    Original copyright notice: Copyright 2013. Derek Bingham, Simon Fraser University.
    """
    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    A = np.array(
        [
            [10, 3, 17, 3.5, 1.7, 8],
            [0.05, 10, 17, 0.1, 8, 14],
            [3, 3.5, 1.7, 10, 17, 8],
            [17, 8, 0.05, 10, 0.1, 14],
        ]
    )
    P = 1e-4 * np.array(
        [
            [1312, 1696, 5569, 124, 8283, 5886],
            [2329, 4135, 8307, 3736, 1004, 9991],
            [2348, 1451, 3522, 2883, 3047, 6650],
            [4047, 8828, 8732, 5743, 1091, 381],
        ]
    )

    outer = 0
    for ii in range(4):
        inner = 0.0
        for jj in range(6):
            xj = x[:, jj]
            Aij = A[ii, jj]
            Pij = P[ii, jj]
            inner = inner + Aij * (xj - Pij) ** 2
        new = alpha[ii] * np.exp(-inner)
        outer = outer + new

    z = -outer

    return z


def borehole(x):
    """Compute the water flow rate through a borehole.

    The Borehole function [1] models water flow through a borehole. Its simplicity
    and quick evaluation makes it a commonly used function for testing a wide variety
    of methods in computer experiments.

    Parameters
    ----------
    x : numpy.ndarray of shape (n_samples, 8)
        The input variables and their usual input ranges:
    
        * rw : radius of borehole in meters, range [0.05, 0.15]
        * r : radius of influence in meters, range [100, 50000]
        * Tu : transmissivity of upper aquifer in m^2/yr, range [63070, 115600]
        * Hu : potentiometric head of upper aquifer in meters, range [990, 1110]
        * Tl : transmissivity of lower aquifer in m^2/yr, range [63.1, 116]
        * Hl : potentiometric head of lower aquifer in meters, range [700, 820]
        * L : length of borehole in meters, range [1120, 1680]
        * Kw : hydraulic conductivity of borehole in m/yr, range [9855, 12045]

    Returns
    -------
    numpy.ndarray of shape (n_samples,)
        The water flow rate in m^3/yr.

    Notes
    -----
    .. [1] Harper, W. V., & Gupta, S. K. (1983). Sensitivity/uncertainty analysis of
        a borehole scenario comparing Latin Hypercube Sampling and deterministic
        sensitivity approaches (No. BMI/ONWI-516). Battelle Memorial Inst., Columbus,
        OH (USA). Office of Nuclear Waste Isolation.

    The distributions of the input random variables are:
    
    * rw ~ N(0.10, 0.0161812)
    * r ~ Lognormal(7.71, 1.0056)
    * Tu ~ Uniform[63070, 115600]
    * Hu ~ Uniform[990, 1110]
    * Tl ~ Uniform[63.1, 116]
    * Hl ~ Uniform[700, 820]
    * L ~ Uniform[1120, 1680]
    * Kw ~ Uniform[9855, 12045]

    Above, N(µ, s^2) is the Normal distribution with mean µ and variance s^2.
    Lognormal(µ, s) is the Lognormal distribution of a variable, such that the natural
    logarithm of the variable has a N(µ, s^2) distribution.

    Authors: Sonja Surjanovic and Derek Bingham, Simon Fraser University

    Original copyright notice: Copyright 2013. Derek Bingham, Simon Fraser University.
    """
    rw = x[:, 0]
    r = x[:, 1]
    Tu = x[:, 2]
    Hu = x[:, 3]
    Tl = x[:, 4]
    Hl = x[:, 5]
    L = x[:, 6]
    Kw = x[:, 7]

    frac1 = 2 * np.pi * Tu * (Hu - Hl)

    frac2a = 2 * L * Tu / (np.log(r / rw) * rw**2 * Kw)
    frac2b = Tu / Tl
    frac2 = np.log(r / rw) * (1 + frac2a + frac2b)

    z = frac1 / frac2

    return z


def detpep8d(x):
    """Dette & Pepelyshev (2010) 8-Dimensional Function

    This function is used for the comparison of computer experiment
    designs. It is highly curved in some variables and less in others
    [1].

    Input Domain:

    The function is evaluated on the hypercube :math:`x_i \in [0, 1], i = 1, \ldots, 8`.

    Parameters
    ----------
    x : numpy.ndarray
        2D array of shape (n, 8) containing n samples with 8 variables each

    Returns
    -------
    numpy.ndarray
        1D array of shape (n,) containing the function values for each input sample

    Notes
    -----
    .. [1] Dette, H., & Pepelyshev, A. (2010). Generalized Latin
        hypercube design for computer experiments. Technometrics,
        52(4).

    Based on https://www.sfu.ca/~ssurjano/detpep108d.html

    Authors: Sonja Surjanovic and Derek Bingham, Simon Fraser University

    Original copyright notice: Copyright 2013. Derek Bingham, Simon Fraser University.
    """
    x1 = x[:, 0]
    x2 = x[:, 1]
    x3 = x[:, 2]

    term1 = 4 * (x1 - 2 + 8 * x2 - 8 * x2**2) ** 2
    term2 = (3 - 4 * x2) ** 2
    term3 = 16 * np.sqrt(x3 + 1) * (2 * x3 - 1) ** 2

    outer = 0
    for ii in range(4, 9):
        inner = 0.0
        for jj in range(3, ii + 1):
            xj = x[:, jj - 1]
            inner = inner + xj

    new = ii * np.log(1 + inner)
    outer = outer + new

    z = term1 + term2 + term3 + outer

    return z



## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------

import numpy as np

def ftos(x, fp=3):
    if x == float('inf'):
        return "+Inf"
    elif x == float('-inf'):
        return "-Inf"
    abs_x = abs(x)
    if x == 0:
        return "0.0"
    elif abs_x >= 0.1 and abs_x < 1000:
        return f"{x:.{fp}f}"
    elif abs_x >= 0.01 and abs_x < 0.1:
        return f"{x:.{fp+1}f}"
    else:
        exponent = int(np.floor(np.log10(abs_x)))
        coeff = x / 10**exponent
        return f"{coeff:.{fp}f}e{exponent}"


class DataFrame:
    def __init__(self, data, colnames, rownames):
        self.data = np.array(data)
        self.rownames = rownames
        self.colnames = colnames
        
    def __getitem__(self, key):
        if isinstance(key, tuple):
            row_key, col_key = key
            if isinstance(row_key, slice) and isinstance(col_key, slice):
                return DataFrame(self.data[row_key, col_key], self.rownames[row_key], self.colnames[col_key])
            elif isinstance(row_key, slice):
                return DataFrame(self.data[row_key, self.colnames.index(col_key)], self.rownames[row_key], [col_key])
            elif isinstance(col_key, slice):
                return DataFrame(self.data[self.rownames.index(row_key), col_key], [row_key], self.colnames[col_key])
            else:
                return self.data[self.rownames.index(row_key), self.colnames.index(col_key)]
        elif isinstance(key, str):
            if key in self.rownames:
                return DataFrame(self.data[self.rownames.index(key), :], [key], self.colnames)
            elif key in self.colnames:
                return DataFrame(self.data[:, self.colnames.index(key)], self.rownames, [key])
            else:
                raise KeyError(f"Key '{key}' not found in row or column names")
        else:
            raise TypeError("Invalid key type. Must be a tuple or a string.")
    
    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            row_key, col_key = key
            if isinstance(row_key, slice) and isinstance(col_key, slice):
                self.data[row_key, col_key] = value
            elif isinstance(row_key, slice):
                self.data[row_key, self.colnames.index(col_key)] = value
            elif isinstance(col_key, slice):
                self.data[self.rownames.index(row_key), col_key] = value
            else:
                self.data[self.rownames.index(row_key), self.colnames.index(col_key)] = value
        elif isinstance(key, str):
            if key in self.rownames:
                self.data[self.rownames.index(key), :] = value
            elif key in self.colnames:
                self.data[:, self.colnames.index(key)] = value
            else:
                raise KeyError(f"Key '{key}' not found in row or column names")
        else:
            raise TypeError("Invalid key type. Must be a tuple or a string.")

    def __repr__(self):
        header = [[''] + self.colnames] 
        rows = header + \
            [[self.rownames[i]+':'] + \
             [ftos(self.data[i, j]) for j in range(self.data.shape[1])] \
             for i in range(self.data.shape[0])]

        min_width = 8
        col_widths = [max(min_width, max(len(str(rows[i][j])) for i in range(len(rows)))) \
                      for j in range(len(rows[0]))]

        formatted_rows = [' '.join(str(rows[i][j]).rjust(col_widths[j]) \
                                   for j in range(len(rows[0]))) \
                          for i in range(len(rows))]

        return '\n'.join(formatted_rows)
        
    def append_row(self, row_data, row_name):
        self.data = np.vstack([self.data, row_data])
        self.rownames.append(row_name)
    
    def append_col(self, col_data, col_name):
        self.data = np.hstack([self.data, np.atleast_2d(col_data).T])
        self.colnames.append(col_name)

    def concat(self, other, axis=0):
        if axis == 0:
            if self.colnames != other.colnames:
                raise ValueError("DataFrames must have the same column names to concatenate vertically")
            new_data = np.concatenate([self.data, other.data], axis=0)
            new_rownames = self.rownames + other.rownames
            return DataFrame(new_data, self.colnames, new_rownames)
        elif axis == 1:
            if self.rownames != other.rownames:
                raise ValueError("DataFrames must have the same row names to concatenate horizontally")
            new_data = np.concatenate([self.data, other.data], axis=1)
            new_colnames = self.colnames + other.colnames
            return DataFrame(new_data, new_colnames, self.rownames)
        else:
            raise ValueError("Axis must be 0 or 1")



"""
Plot and optimize the restricted negative log-likelihood

Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2023, CentraleSupelec
License: GPLv3 (see LICENSE)
"""

import gpmp.num as gnp
import gpmp as gp
import matplotlib.pyplot as plt


def f_matern(x):
    l = 0.1  # length scale
    x1, x2, x3, x4 = -0.5, -0.2, 0.3, 0.5  # translation points
    x = gnp.asarray(x)
    return gnp.exp(-1.0 + 0.2 * x**2 - 
        0.1 * gp.kernel.matern32_kernel(gnp.abs(x - x1) / 0.01)
        - 0.2 * gp.kernel.matern32_kernel(gnp.abs(x - x2) / 0.02)
        - 0.1 * gp.kernel.matern32_kernel(gnp.abs(x - x3) / 0.04)
        - 0.2 * gp.kernel.matern32_kernel(gnp.abs(x - x4) / 0.04)
    )


def generate_data():
    """
    Data generation.

    Returns
    -------
    tuple
        (xt, zt): target data
        (xi, zi): input dataset
    """
    dim = 1
    nt = 1000
    box = [[-1], [1]]
    xt = gp.misc.designs.regulargrid(dim, nt, box)
    zt = f_matern(xt)

    ni = 40
    xi = gp.misc.designs.ldrandunif(dim, ni, box)
    zi = f_matern(xi)

    return xt, zt, xi, zi


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def kernel(x, y, covparam, pairwise=False):
    p = 1
    return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)


def visualize_results(xt, zt, xi, zi, zpm, zpv):
    """
    Visualize the results using gp.misc.plotutils (a matplotlib wrapper).

    Parameters
    ----------
    xt : numpy.ndarray
        Target x values
    zt : numpy.ndarray
        Target z values
    xi : numpy.ndarray
        Input x values
    zi : numpy.ndarray
        Input z values
    zpm : numpy.ndarray
        Posterior mean
    zpv : numpy.ndarray
        Posterior variance
    """
    fig = gp.misc.plotutils.Figure(isinteractive=True)
    fig.plot(xt, zt, "k", linewidth=1, linestyle=(0, (5, 5)))
    fig.plotdata(xi, zi)
    fig.plotgp(xt, zpm, zpv, colorscheme="simple")
    fig.xylabels("$x$", "$z$")
    fig.title("Posterior GP with parameters selected by ReML")
    fig.show(grid=True, xlim=[-1.0, 1.0], legend=True, legend_fontsize=9)


def main():
    xt, zt, xi, zi = generate_data()

    model = gp.core.Model(constant_mean, kernel)

    # Automatic selection of parameters using REML
    model, info = gp.kernel.select_parameters_with_remap(model, xi, zi, info=True)
    gp.misc.modeldiagnosis.diag(model, info, xi, zi)

    # Prediction
    zpm, zpv = model.predict(xi, zi, xt)

    # Visualization
    print("\nVisualization")
    print("-------------")
    plot_likelihood = False
    if plot_likelihood:
        gp.misc.modeldiagnosis.plot_likelihood_sigma_rho(model, info)

    visualize_results(xt, zt, xi, zi, zpm, zpv)

    zloo, sigma2loo, eloo = model.loo(xi, zi)

    plt.figure()
    plt.plot(zi, eloo / sigma2loo, 'o')
    plt.show()

    r = gnp.to_np(eloo / sigma2loo)
    
    from scipy.stats import gennorm
    import numpy as np
    
    shape, loc, scale = gennorm.fit(r)
    print(f"Shape (beta): {shape}")
    print(f"Location (mu): {loc}")
    print(f"Scale (alpha): {scale}")

    # Plot the histogram of the data
    plt.figure()
    plt.hist(r, bins=30, density=True, alpha=0.6, color='g')

    # Create a range of values for the PDF
    xmin, xmax = plt.xlim()
    x_range = np.linspace(xmin, xmax, 100)

    # Plot the fitted PDF
    pdf_fitted = gennorm.pdf(x_range, shape, loc, scale)
    plt.plot(x_range, pdf_fitted, 'k', linewidth=2)

    # Display the plot
    plt.title('Fit of Generalized Gaussian Distribution')
    plt.xlabel('Data')
    plt.ylabel('Density')
    plt.show()


if __name__ == "__main__":
    main()
